#pragma once

#include <base/heap.h>
#include <base/log.h>
#include <os/reporter.h>
#include <qumela/types.h>

namespace Qumela {

	using namespace Genode;

	class Generatible;
	
	template <class T>
	class Element : public Fifo<Element<T>>::Element
	{
	private:
		T* _value;

	public:
		T* value() const { return _value; }

		Element(T* value)
			: _value(value)
		{
		}
	};
	
	class Generatible
	{
	public:
		virtual void generate(Xml_generator &xml) = 0;

		virtual ~Generatible() { }

		Generatible() { }
	};

	template <class T>
	class Generatible_container : public virtual Generatible
	{
	private:
		Fifo<Element<T>> _children { };

	protected:
		Heap &_heap;

		void add(T* g)
		{
			_children.enqueue(new (_heap) Element<T>(g));
		}

		bool has_children()
		{
			return !_children.empty();
		}

		void generate_children(Xml_generator &xml)
		{
			for (Element<T>* e = _children.head(); e != nullptr; e = e->next())
			{
				e->value()->generate(xml);
			}
		}

	public:
		virtual ~Generatible_container()
		{
			while (!_children.empty())
			{
				auto c = _children.dequeue();
				destroy(_heap, c->value());
				destroy(_heap, c);
			}
		}

		Generatible_container(Heap &heap)
			: _heap(heap)
		{
		}
	};	
	
	class Vfs_entry : public virtual Generatible
	{
	};
	
	class Vfs_node : public Vfs_entry
	{
	private:
		const Vfs_name _name;

	public:
		virtual void generate(Xml_generator &xml) override
		{
			xml.node(_name.string(), [&] () { });
		}
		
		Vfs_node(const Vfs_name& name)
			: _name(name)
		{
		}
	};
	
	class Vfs_inline : public Vfs_entry
	{
	private:
		const Vfs_name _name;
		const String<4096> _content;

	public:
		virtual void generate(Xml_generator &xml) override
		{
			xml.node("inline", [&] () { 
				xml.attribute("name", _name);
				xml.append(_content.string());
			});
		}
		
		Vfs_inline(const Vfs_name& name, const Vfs_content& content)
			: _name(name), _content(content)
		{
		}
	};
	
	class Vfs_fs : public Vfs_entry
	{
	private:
		const Interface_label _label;

	public:
		virtual void generate(Xml_generator &xml) override
		{
			xml.node("fs", [&] () { 
				if (_label != "") {
					xml.attribute("label", _label);
				}
			} );
		}
		
		Vfs_fs(const Interface_label& label)
			: _label(label)
		{
		}
	};
	
	class Vfs_dir : public Vfs_entry, public Generatible_container<Vfs_entry>
	{
	private:
		const Vfs_name _name;

	public:
		Vfs_node* node(const Vfs_name& name)
		{
			Vfs_node* n = new (_heap) Vfs_node(name);
			add(n);
			return n;
		}

		Vfs_dir* dir(const Vfs_name& name)
		{
			Vfs_dir* d = new (_heap) Vfs_dir(_heap, name);
			add(d);
			return d;
		}

		Vfs_fs* fs(const Interface_label& label)
		{
			Vfs_fs* f = new (_heap) Vfs_fs(label);
			add(f);
			return f;
		}
		
		Vfs_fs* fs()
		{
			return fs("");
		}

		Vfs_inline* file(const Vfs_name& name, const Vfs_content& content)
		{
			Vfs_inline* i = new (_heap) Vfs_inline(name, content);
			add(i);
			return i;
		}
		
		virtual void generate(Xml_generator &xml) override
		{
			xml.node("dir", [&] () { 
				xml.attribute("name", _name);
				generate_children(xml);
			} );
		}
		
		Vfs_dir(Heap &heap, const Vfs_name& name)
			: Generatible_container(heap), _name(name)
		{
		}
	};
		
	class Service_route : public virtual Generatible
	{
	};
	
	class Vfs : public Generatible_container<Vfs_entry>
	{
	public:
		Vfs_dir* dir(const Vfs_name& name)
		{
			Vfs_dir* d = new (_heap) Vfs_dir(_heap, name);
			add(d);
			return d;
		}

		Vfs_fs* fs(const Interface_label& label)
		{
			Vfs_fs* f = new (_heap) Vfs_fs(label);
			add(f);
			return f;
		}
		
		Vfs_fs* fs()
		{
			return fs("");
		}
		
		Vfs_inline* file(const Vfs_name& name, const Vfs_content& content)
		{
			Vfs_inline* i = new (_heap) Vfs_inline(name, content);
			add(i);
			return i;
		}

		virtual void generate(Xml_generator &xml) override
		{
			xml.node("vfs", [&] () { 
				generate_children(xml);
			} );
		}

		Vfs(Heap &heap)
			: Generatible_container(heap)
		{
		}
	};
	
	class Fetch_url : public virtual Generatible
	{
	private:
		const Url _url;
		const File_path _path;

	public:
		virtual void generate(Xml_generator &xml) override
		{
			xml.node("fetch", [&] () { 
				xml.attribute("url", _url);
				xml.attribute("path", _path);
			} );
		}
		
		Fetch_url(const Url& url, const File_path& path)
			: _url(url), _path(path)
		{
		}
	};
	
	class Verify_signature : public virtual Generatible
	{
	private:
		const File_path _path;
		const File_path _pubkey;

	public:
		virtual void generate(Xml_generator &xml) override
		{
			xml.node("verify", [&] () { 
				xml.attribute("path", _path);
				xml.attribute("pubkey", _pubkey);
			} );
		}
		
		Verify_signature(const File_path& path, const File_path& pubkey)
			: _path(path), _pubkey(pubkey)
		{
		}
	};
	
	class Start_report : public virtual Generatible
	{
	private:
		const bool _progress;

	public:
		virtual void generate(Xml_generator &xml) override
		{
			xml.node("report", [&] () { 
				xml.attribute("progress", _progress ? "yes" : "no");
			} );
		}
		
		Start_report(const bool progress)
			: _progress(progress)
		{
		}
	};
	
	class Default_policy : public virtual Generatible
	{
	private:
		const File_path _path;
		const bool _writeable;

	public:
		virtual void generate(Xml_generator &xml) override
		{
			xml.node("default-policy", [&] () { 
				xml.attribute("path", _path);
				xml.attribute("writeable", _writeable ? "yes" : "no");
			} );
		}
		
		Default_policy(const File_path& path, const bool writeable)
			: _path(path), _writeable(writeable)
		{
		}
	};
	
	class Child : public Service_route
	{
	private:
		const Component_name _name;
		const Interface_label _label;

	public:
		virtual void generate(Xml_generator &xml) override
		{
			xml.node("child", [&] () { 
				xml.attribute("name", _name);
				if (_label != "") {
					xml.attribute("label", _label);
				}
			} );
		}
		
		Child(const Component_name& name, const Interface_label& label)
			: _name(name), _label(label)
		{
		}
	};
	
	class Parent : public Service_route
	{
	private:
		const Interface_label _label;

	public:
		virtual void generate(Xml_generator &xml) override
		{
			xml.node("parent", [&] () { 
				if (_label != "") {
					xml.attribute("label", _label);
				}
			} );
		}
		
		Parent(const Interface_label& label)
			: _label(label)
		{
		}
	};
	
	class Service : public Generatible_container<Service_route>
	{
	private:
		const Interface_type _name;
		const Interface_label _label;
		const Interface_label _label_last;

	public:
		virtual void generate(Xml_generator &xml) override
		{
			xml.node("service", [&] () {
				xml.attribute("name", _name);

				if (_label != "") {
					xml.attribute("label", _label);
				}
				
				if (_label_last != "") {
					xml.attribute("label_last", _label_last);
				}

				generate_children(xml);
			} );
		}
		
		Parent* parent()
		{
			Parent* p = new (_heap) Parent("");
			add(p);
			return p;
		}
		
		Parent* parent(const Interface_label& label)
		{
			Parent* p = new (_heap) Parent(label);
			add(p);
			return p;
		}
		
		Child* child(const Component_name& name)
		{
			Child* c = new (_heap) Child(name, "");
			add(c);
			return c;
		}
		
		Child* child(const Component_name& name, const Interface_label& label)
		{
			Child* c = new (_heap) Child(name, label);
			add(c);
			return c;
		}
		
		Service(Heap &heap, const Component_name& name, const Interface_label& label, const Interface_label& label_last)
			: Generatible_container(heap), _name(name), _label(label), _label_last(label_last)
		{
		}
	};
	
	class Libc : public virtual Generatible
	{
	private:
		const File_path _stdout;
		const File_path _stderr;
		const File_path _rtc;
		const File_path _socket;

	public:
		virtual void generate(Xml_generator &xml) override
		{
			xml.node("libc", [&] () {
				if (_stdout != "") { 
					xml.attribute("stdout", _stdout);
				}
				if (_stderr != "") { 
					xml.attribute("stderr", _stderr);
				}
				if (_rtc != "") { 
					xml.attribute("rtc", _rtc);
				}
				if (_socket != "") { 
					xml.attribute("socket", _socket);
				}
			} );
		}
		
		Libc(const File_path& stdout_, const File_path& stderr_, const File_path& rtc, const File_path& socket)
		  : _stdout(stdout_), _stderr(stderr_), _rtc(rtc), _socket(socket)
		{
		}
	};
	
	class Config_xml : public virtual Generatible
	{
	};
	
	class Config_attribute : public Config_xml
	{
	private:
		const Attribute_name _name;
		const Attribute_value _value;

	public:
		virtual void generate(Xml_generator &xml) override
		{
			xml.attribute(_name.string(), _value.string());
		}
		
		Config_attribute(const Attribute_name& name, const Attribute_value& value)
			: _name(name), _value(value)
		{
		}
	};
	
	class Config_node : public Config_xml, public Generatible_container<Config_xml>
	{
	private:
		const Node_name _name;

	public:
		virtual void generate(Xml_generator &xml) override
		{
			xml.node(_name.string(), [&] () {
				generate_children(xml);
			} );
		}
		
		Config_node* node(const Node_name& name)
		{
			Config_node* n = new (_heap) Config_node(_heap, name);
			add(n);
			return n;
		}
		
		Config_attribute* attribute(const Attribute_name& name, const Attribute_value& value)
		{
			Config_attribute* a = new (_heap) Config_attribute(name, value);
			add(a);
			return a;
		}
		
		Config_node(Heap &heap, Node_name name)
			: Generatible_container(heap), _name(name)
		{
		}
	};
	
	class Start_config : public Generatible_container<Generatible>
	{
	private:
		Vfs* _vfs = nullptr;
		bool _verbose = false;

	public:
		Start_config(const Start_config&) = delete;
		Start_config& operator=(const Start_config&) = delete;

		bool verbose() const { return _verbose; }
		void verbose(bool v) { _verbose = v; }

		virtual void generate(Xml_generator &xml) override
		{
			xml.node("config", [&] () {
				xml.attribute("verbose", _verbose ? "yes" : "no");
				generate_children(xml);
			});
		}

		Vfs* vfs()
		{
			if (_vfs == nullptr) {
				_vfs = new (_heap) Vfs(_heap);
				add(_vfs);
			}

			return _vfs;
		}

		Default_policy* default_policy(const File_path& path, const bool writeable)
		{
			Default_policy* p = new (_heap) Default_policy(path, writeable);
			add(p);
			return p;
		}
		
		Libc* libc(const File_path& stdout_, const File_path& stderr_, const File_path& rtc, const File_path& socket)
		{
			Libc* l = new (_heap) Libc(stdout_, stderr_, rtc, socket);
			add(l);
			return l;
		}
		
		Start_report* report(const bool progress)
		{
			Start_report* p = new (_heap) Start_report(progress);
			add(p);
			return p;
		}
		
		Fetch_url* fetch_url(const Url& url, const File_path& path)
		{
			Fetch_url* f = new (_heap) Fetch_url(url, path);
			add(f);
			return f;
		}

		Verify_signature* verify_signature(const File_path& path, const File_path& pubkey)
		{
			Verify_signature* v = new (_heap) Verify_signature(path, pubkey);
			add(v);
			return v;
		}
		
		Config_node* node(const Node_name& name)
		{
			Config_node* n = new (_heap) Config_node(_heap, name);
			add(n);
			return n;
		}
		
		Config_attribute* attribute(const Attribute_name& name, const Attribute_value& value)
		{
			Config_attribute* a = new (_heap) Config_attribute(name, value);
			add(a);
			return a;
		}
		
		Start_config(Heap &heap)
			: Generatible_container(heap)
		{
		}
	};
	
	class Parent_provides : public Generatible_container<Service>
	{
	public:
		virtual void generate(Xml_generator &xml) override
		{
			if (has_children()) {
				xml.node("parent-provides", [&] () {
					generate_children(xml);
				});
			}
		}
		
		Service* service(const Component_name& name)
		{
			Service* s = new (_heap) Service(_heap, name, "", "");
			add(s);
			return s;
		}
		
		Parent_provides(Heap &heap)
			: Generatible_container(heap)
		{
		}
	};
	
	class Provides : public Generatible_container<Service>
	{
	public:
		virtual void generate(Xml_generator &xml) override
		{
			if (has_children()) {
				xml.node("provides", [&] () {
					generate_children(xml);
				});
			}
		}
		
		Service* service(const Component_name& name)
		{
			Service* s = new (_heap) Service(_heap, name, "", "");
			add(s);
			return s;
		}
		
		Provides(Heap &heap)
			: Generatible_container(heap)
		{
		}
	};
	
	class Route : public Generatible_container<Service>
	{
	public:
		virtual void generate(Xml_generator &xml) override
		{
			if (has_children()) {
				xml.node("route", [&] () {
					generate_children(xml);
				});
			}
		}
		
		Service* service(const Component_name& name)
		{
			Service* s = new (_heap) Service(_heap, name, "", "");
			add(s);
			return s;
		}

		Service* service(const Component_name& name, const Interface_label& label)
		{
			Service* s = new (_heap) Service(_heap, name, label, "");
			add(s);
			return s;
		}
		
		Service* service(const Component_name& name, const Interface_label& label, const Interface_label& label_last)
		{
			Service* s = new (_heap) Service(_heap, name, label, label_last);
			add(s);
			return s;
		}
		
		Route(Heap &heap)
			: Generatible_container(heap)
		{
		}
	};
	
	class Common_routes : public Generatible_container<Service>
	{
	public:
		virtual void generate(Xml_generator &xml) override
		{
			if (has_children()) {
				xml.node("common_routes", [&] () {
					generate_children(xml);
				});
			}
		}
		
		Service* service(const Component_name& name)
		{
			Service* s = new (_heap) Service(_heap, name, "", "");
			add(s);
			return s;
		}

		Service* service(const Component_name& name, const Interface_label& label)
		{
			Service* s = new (_heap) Service(_heap, name, label, "");
			add(s);
			return s;
		}
		
		Service* service(const Component_name& name, const Interface_label& label, const Interface_label& label_last)
		{
			Service* s = new (_heap) Service(_heap, name, label, label_last);
			add(s);
			return s;
		}
		
		Common_routes(Heap &heap)
			: Generatible_container(heap)
		{
		}
	};

	class Init_start : public Generatible_container<Generatible>
	{
	private:
		const Component_name _name;
		const Component_name _binary;
		const size_t _caps = 0;
		const Number_of_bytes _ram = 0;
		Provides* _provides = nullptr;
		Route* _route = nullptr;
		Start_config* _config = nullptr;

	public:
		Init_start(const Init_start&) = delete;
		Init_start& operator=(const Init_start&) = delete;
		
		Init_start(Heap &heap, const Component_name& name, const Component_name& binary, const size_t caps, const Number_of_bytes ram)
			: Generatible_container(heap), _name(name),
				_binary(binary), _caps(caps), _ram(ram)
		{
		}

		virtual void generate(Xml_generator &xml) override
		{
			xml.node("start", [&] () {
				xml.attribute("name", _name);
				if (_caps > 0) {
					xml.attribute("caps", _caps);
				}
				if (_ram > 0) {
					xml.node("resource", [&] () {
						xml.attribute("name", "RAM");
						xml.attribute("quantum", _ram);
					});
				}
				if (_binary != "") {
					xml.node("binary", [&] () {
						xml.attribute("name", _binary);
					});
				}
				generate_children(xml);
			});
		}

		Route* route()
		{
			if (_route == nullptr)
			{
				_route = new (_heap) Route(_heap);
				add(_route);
			}

			return _route;
		}
		
		Start_config* config()
		{
			if (_config == nullptr)
			{
				_config = new (_heap) Start_config(_heap);
				add(_config);
			}

			return _config;
		}
		
		Provides* provides()
		{
			if (_provides == nullptr)
			{
				_provides = new (_heap) Provides(_heap);
				add(_provides);
			}

			return _provides;
		}
	};

	class Deploy_start : public Generatible_container<Generatible>
	{
	private:
		const Component_name _name;
		const Package_path _package;
		const Number_of_bytes _ram = 0;
		Route* _route = nullptr;
		Start_config* _config = nullptr;

	public:
		Deploy_start(const Deploy_start&) = delete;
		Deploy_start& operator=(const Deploy_start&) = delete;

		Deploy_start(Heap &heap, const Component_name& name, const Package_path& package)
		: Generatible_container(heap), _name(name), _package(package)
		{
		}

		Deploy_start(Heap &heap, const Component_name& name, const Package_path& package, const Number_of_bytes ram)
		: Generatible_container(heap), _name(name), _package(package), _ram(ram)
		{
		}

		virtual void generate(Xml_generator &xml) override
		{
			xml.node("start", [&] () {
				xml.attribute("name", _name);
				xml.attribute("pkg", _package);
				if (_ram > 0) {
					xml.attribute("ram", _ram);
				}
				generate_children(xml);
			});
		}

		Route* route()
		{
			if (_route == nullptr)
			{
				_route = new (_heap) Route(_heap);
				add(_route);
			}

			return _route;
		}

		Start_config* config()
		{
			if (_config == nullptr)
			{
				_config = new (_heap) Start_config(_heap);
				add(_config);
			}

			return _config;
		}
	};

	class Init_report : public virtual Generatible
	{
	private:
		bool _init_ram = false;
		bool _init_caps = false;
		bool _child_ram = false;
		bool _child_caps = false;
		size_t _delay_ms = 1000;
		Number_of_bytes _buffer_size = 64 * 1024;

	public:
		bool init_ram() { return _init_ram; }
		void init_ram(bool value) { _init_ram = value; }
		bool init_caps() { return _init_caps; }
		void init_caps(bool value) { _init_caps = value; }
		bool child_ram() { return _child_ram; }
		void child_ram(bool value) { _child_ram = value; }
		bool child_caps() { return _child_caps; }
		void child_caps(bool value) { _child_caps = value; }
		size_t delay_ms () { return _delay_ms; }
		void delay_ms(bool value) { _delay_ms = value; }
		Number_of_bytes buffer_size() { return _buffer_size; }
		void buffer_size(bool value) { _buffer_size = value; }

		virtual void generate(Xml_generator &xml) override
		{
			xml.node("report", [&] () { 
				xml.attribute("init_ram", _init_ram ? "yes" : "no");
				xml.attribute("init_caps", _init_caps ? "yes" : "no");
				xml.attribute("child_ram", _child_ram ? "yes" : "no");
				xml.attribute("child_caps", _child_caps ? "yes" : "no");
				xml.attribute("delay_ms", _delay_ms);
				xml.attribute("buffer", _buffer_size);
			} );
		}
	};

	class Init_config : public Generatible_container<Generatible>
	{
	private:
		Common_routes* _common_routes = nullptr;
		Parent_provides* _parent_provides = nullptr;
		Init_report* _report = nullptr;
	
	public:
		Init_config(const Init_config&) = delete;
		Init_config& operator=(const Init_config&) = delete;

		Init_config(Heap &heap)
			: Generatible_container(heap)
		{
		}

		virtual void generate(Xml_generator &xml) override
		{
			xml.attribute("verbose", "yes");
			generate_children(xml);
		}
		
		Init_report* report()
		{
			if (_report == nullptr) {
				_report = new (_heap) Init_report();
				add(_report);
			}

			return _report;
		}

		Parent_provides* parent_provides()
		{
			if (_parent_provides == nullptr) {
				_parent_provides = new (_heap) Parent_provides(_heap);
				add(_parent_provides);
			}

			return _parent_provides;
		}

		Init_start* start(const Component_name& name, const Component_name& binary, const size_t caps, const Number_of_bytes ram)
		{
			Init_start* s = new (_heap) Init_start(_heap, name, binary, caps, ram);
			add(s);
			return s;
		}
	};

	class Deploy_config : public Generatible_container<Generatible>
	{
	private:
		Common_routes* _common_routes = nullptr;

	public:
		Deploy_config(const Deploy_config&) = delete;
		Deploy_config& operator=(const Deploy_config&) = delete;

		Deploy_config(Heap &heap)
		: Generatible_container(heap)
		{
		}

		virtual void generate(Xml_generator &xml) override
		{
			xml.attribute("arch", "x86_64");
			generate_children(xml);
		}

		Common_routes* common_routes()
		{
			if (_common_routes == nullptr) {
				_common_routes = new (_heap) Common_routes(_heap);
				add(_common_routes);
			}

			return _common_routes;
		}

		Deploy_start* start(const Component_name& name, const Package_path& package)
		{
			Deploy_start* s = new (_heap) Deploy_start(_heap, name, package);
			add(s);
			return s;
		}

		Deploy_start* start(const Component_name& name, const Package_path& package, const Number_of_bytes ram)
		{
			Deploy_start* s = new (_heap) Deploy_start(_heap, name, package, ram);
			add(s);
			return s;
		}

	};

}
